/**
 * @file canethlib.cpp
 *
 * CAN C/C++ API for proconX CAN-ETH gateway
 *
 * @if NOTICE
 *
 * Copyright (c) 2011-2020 proconX Pty Ltd. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PROCONX AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL PROCONX OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For additional information see http://www.proconx.com
 *
 * @endif
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#ifdef _WIN32
#   define WIN32_LEAN_AND_MEAN
#   include <windows.h>
#   include <winsock.h>
#   pragma comment(lib, "Ws2_32.lib")
#else
#   include <unistd.h> 
#   include <sys/socket.h>
#   include <netinet/in.h>
#   include <arpa/inet.h>
#   include <netdb.h>
#   include <pthread.h>
#endif
#ifdef TRACE_TOOL
#   include "tracetool\tracetool.h"
#   define TRACE_SEND(watchName, watchValue) TTrace::Debug()->Send(watchName, watchValue)
#else
#   define TRACE_SEND(watchName, watchValue)
#endif
#include "canethlib.h"


/*****************************************************************************
 * Definitions
 *****************************************************************************/

#define MAX_CAN_GATEWAYS               32 ///< Number of gateways supported

#define CANETH_FRAME_VERSION           2
#define MAX_CAN_MSG                    96 // v1 has up to 16 messages, v2 has up to 96
#define UDP_PORT                       11898
const char MAGIC_ISO11898_ID[] = {'I', 'S', 'O', '1', '1', '8', '9', '8'};


/*****************************************************************************
 * Some cross-platform definitions
 *****************************************************************************/

#ifdef _WIN32
#else   
#   ifndef INVALID_SOCKET
#       define INVALID_SOCKET ((SOCKET)(~0))
#   endif
#   define closesocket(sock) close(sock)
    typedef int SOCKET;
#endif

#if defined(HAVE_SOCKLEN_T) || defined(__socklen_t_defined) || defined(socklen_t) || \
    defined(_SOCKLEN_T_DECLARED) || defined(_SOCKLEN_T)
   typedef socklen_t SOCK_LEN_TYPE;
#elif defined(HAVE_SOCKLEN_U)
   typedef unsigned int SOCK_LEN_TYPE;
#else
   typedef int SOCK_LEN_TYPE;
#endif


/*****************************************************************************
 * Types
 *****************************************************************************/

/*
 * Structure of a CAN-ETH variable length UDP message.
 *
 * @note Has to be packed to byte boundary!
 */
#pragma pack(1)
typedef struct
{
    char magicId[8]; // Must be 'ISO11898'
    unsigned char version; // version 1 and 2 are supported
    unsigned char cnt; // v1 has up to 16 messages, v2 has up to 96
    CanMsg messages[MAX_CAN_MSG];
    char options[32]; // v1 has up to 128 option bytes, v2 has up to 32 only
} CanEthFrame;
#pragma pack()


class CanBuffer
{   
#ifdef _WIN32
    CRITICAL_SECTION criticalSection;
#else    
    pthread_mutex_t mutex;
#endif    

  public:

    static const int CAN_BUF_SIZE = 128;

    CANFRAME data[CAN_BUF_SIZE];
    unsigned int size;
    unsigned int readIdx;
    unsigned int writeIdx;
    sockaddr_in gatewayIpAddr;
    void (*recvNotifyFunc) (CanHandle hdl);


    void init()
    {
        size = 0;
        readIdx = 0;
        writeIdx = 0;
        recvNotifyFunc = NULL;
#ifdef _WIN32      
        InitializeCriticalSection(&criticalSection);
#else
        pthread_mutex_init(&mutex, NULL);
#endif
    }


   void destroy()
   {
        size = 0;
#ifdef _WIN32      
        DeleteCriticalSection(&criticalSection);
#else      
        pthread_mutex_destroy(&mutex);
#endif       
    }   


    void lock()
    {
#ifdef _WIN32      
        EnterCriticalSection(&criticalSection);
#else
        pthread_mutex_lock(&mutex);
#endif      
   }


    void unlock()
    {
#ifdef _WIN32      
        LeaveCriticalSection(&criticalSection);
#else
        pthread_mutex_unlock(&mutex);
#endif      
    }
};


/*****************************************************************************
 * Data
 *****************************************************************************/

#ifdef _WIN32
    static HANDLE recvThreadHdl;
#else
    static pthread_t recvThreadHdl;
#endif
static SOCKET recvSocket = INVALID_SOCKET;
static SOCKET sendSocket = INVALID_SOCKET;

static CanBuffer canBufArr[MAX_CAN_GATEWAYS];


/*****************************************************************************
 * Helpers
 *****************************************************************************/

/**
 * @internal
 * Init CAN-ETH UDP frame
 *
 * @param frame Pointer to CAN-ETH UDP frame
 *
 * @return int Length in bytes of CAN-ETH UDP frame
 */
int canEthFrameInit(CanEthFrame *frame)
{
   const char CANETH_MAGIC_ID[8] = {'I', 'S', 'O', '1', '1', '8', '9', '8'};

   memmove(&frame->magicId, CANETH_MAGIC_ID, sizeof(frame->magicId));
   frame->version = CANETH_FRAME_VERSION;
   frame->cnt = 0;
   frame->messages[0].id = 0xFFFFFFFF; // Set option mark to EOF

   return 10 + 1; // plus one option byte
}


/**
 * @internal
 * Pack a CAN message into a CAN-ETH UDP frame
 *
 * @param frame Pointer to CAN-ETH UDP frame
 * @param msg Pointer to a CAN message
 *
 * @return int Length in bytes of CAN-ETH UDP frame
 */
int canEthFramePack(CanEthFrame *frame, const CanMsg *const msg)
{
   if (frame->cnt < MAX_CAN_MSG)
   {
#if defined(__BYTE_ORDER) && defined(__BIG_ENDIAN) && __BYTE_ORDER == __BIG_ENDIAN
      // Swap ID bytes in case we run on a big-endian machine
      msg->id = le32toh(msg->id);
#endif
      memcpy(&frame->messages[frame->cnt], msg, sizeof(CanMsg));
      frame->cnt++;
      if (frame->cnt < MAX_CAN_MSG)
          frame->messages[frame->cnt].id = 0xFFFFFFFF; // Set option mark to EOF
      else
          frame->options[0] = (char) 0xFF; // EOF to mark no options contained in this frame
      return sizeof(CanEthFrame) - sizeof(CanMsg) * (MAX_CAN_MSG - frame->cnt);
   }
   else
      return -1;
}


/**
 * @internal
 * Validates a received CAN-ETH UDP frame
 *
 * @param frame Pointer to CAN-ETH UDP frame
 * @param len Number of bytes received
 *
 * @return int True if frame validates OK
 */
int canethFrameValidate(CanEthFrame *frame, int len)
{
    const char MAGIC_ISO11898_ID[8] = {'I', 'S', 'O', '1', '1', '8', '9', '8'};

    // Check minimum length
    if (len <= 10)
        return 0;
    // Check for magic ID
    if (memcmp(&frame->magicId, MAGIC_ISO11898_ID, sizeof(MAGIC_ISO11898_ID)) != 0)
        return 0;
    // Check for supported version
    if (frame->version != 1 && frame->version != 2)
        return 0;
    // Check count within bounds
    if (frame->cnt > MAX_CAN_MSG)
        return 0;

#if defined(__BYTE_ORDER) && defined(__BIG_ENDIAN) && __BYTE_ORDER == __BIG_ENDIAN
    // Swap ID bytes in case we run on a big-endian machine
    for (int i = 0; i < frame->cnt; i++)
        frame->messages[i].id = htole32(frame->messages[i].id);
#endif

    // All tests passed, OK
    return 1;
}


/*****************************************************************************
 * Background threads
 *****************************************************************************/

/**
 * @internal
 * CAN background receive thread
 */
#if defined(_WIN32)
static DWORD WINAPI canRecvThread(LPVOID)
#else
static void* canRecvThread(void *)
#endif
{
    for (;;)
    {
        int len;
        sockaddr_in senderAddr;
        SOCK_LEN_TYPE senderAddrSize;
        CanEthFrame frame;

        senderAddrSize = sizeof(senderAddr);
        len = recvfrom(recvSocket, (char *) &frame, sizeof(frame), 0, (struct sockaddr *) &senderAddr, &senderAddrSize);

        // Correct frame ?
        if (canethFrameValidate(&frame, len))
        {
            int hdl;

            //
            // Search for correct communication endpoint
            //
            for (hdl = 0; hdl < MAX_CAN_GATEWAYS; hdl++)
            {
                int i = 0;
                void (*cb) (CanHandle hdl);

                canBufArr[hdl].lock();
                while ((i < (int) frame.cnt) &&
                    (canBufArr[hdl].size > 0) &&
                    (canBufArr[hdl].gatewayIpAddr.sin_family == senderAddr.sin_family) &&
                    (canBufArr[hdl].gatewayIpAddr.sin_addr.s_addr == senderAddr.sin_addr.s_addr))
                {
#ifdef TRACE_TOOL
                    char buf[1000];
                    sprintf(buf, "%d/%d: dlc=%d ext=%d rtr=%d %lX:%02X %02X %02X %02X %02X %02X %02X %02X",
                        i, (int) frame.cnt,
                        (int) frame.canFrameArr[i].len,
                            (int) frame.canFrameArr[i].ext, (int) frame.canFrameArr[i].rtr, frame.canFrameArr[i].id,
                            (unsigned int) frame.canFrameArr[i].byte[0], (unsigned int) frame.canFrameArr[i].byte[1],
                            (unsigned int) frame.canFrameArr[i].byte[2], (unsigned int) frame.canFrameArr[i].byte[3],
                            (unsigned int) frame.canFrameArr[i].byte[4], (unsigned int) frame.canFrameArr[i].byte[5],
                            (unsigned int) frame.canFrameArr[i].byte[6], (unsigned int) frame.canFrameArr[i].byte[7]);
                    TTrace::Debug()->Send("canRecvThread", buf);
#endif
                    if (canBufArr[hdl].writeIdx == (canBufArr[hdl].readIdx + 1) % canBufArr[hdl].size)
                    {
                        // Buffer overflow, also advance read index
                        canBufArr[hdl].readIdx = (canBufArr[hdl].readIdx + 1) % canBufArr[hdl].size;
                    }
                    memcpy(&canBufArr[hdl].data[canBufArr[hdl].writeIdx], &frame.messages[i], sizeof(canBufArr[hdl].data[canBufArr[hdl].writeIdx]));
                    canBufArr[hdl].writeIdx = (canBufArr[hdl].writeIdx + 1) % canBufArr[hdl].size;
                    cb = canBufArr[hdl].recvNotifyFunc;
                    canBufArr[hdl].unlock();
                    if (cb != NULL)
                        cb(hdl);
                    i++;
                    if (i >= (int) frame.cnt) // Leave inner loop? Then jump out of outer loop too...
                        goto done; // No need to do re-enter the critical section
                    canBufArr[hdl].lock();
                }
                canBufArr[hdl].unlock();
            } // for hdl
        done: ;
        }
        else
        {
            if (len < 0)
            {
                TRACE_SEND("canRecvThread", "recvfrom failed");
            }
            continue; // Discard frame
        }
    }
    return 0;
}


/*****************************************************************************
 * API functions
 *****************************************************************************/

/**
 * Initialises the CAN-ETH Ethernet interface.
 *
 * A background receive thread is created and started.
 *
 * @return CANETH_SUCCESS or a negative error code
 */
int canEthInit()
{
    int rc;
    sockaddr_in recvAddr;
#ifdef _WIN32
    WSADATA wsaData;
    WSAStartup(0x0101, &wsaData); // Enable Windows' socket subsystem
#endif

    //
    // Create a receiver socket to receive datagrams
    //
    recvSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (recvSocket == INVALID_SOCKET)
    {
        TRACE_SEND("canEthInit", "CANETH_SOCKET_FAILED");
        return CANETH_SOCKET_FAILED;
    }

    //
    // Bind the socket to any address and the specified port
    //
    recvAddr.sin_family = AF_INET;
    recvAddr.sin_port = htons(UDP_PORT);
    recvAddr.sin_addr.s_addr = htonl(INADDR_ANY);

    rc = bind(recvSocket, (struct sockaddr *) & recvAddr, sizeof (recvAddr));
    if (rc != 0)
    {
        TRACE_SEND("canEthInit", "CANETH_PORT_ALREADY_BOUND");
        closesocket(recvSocket);
        return CANETH_PORT_ALREADY_BOUND;
    }

    //
    // Create a socket for sending data
    //
    sendSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (sendSocket == INVALID_SOCKET)
    {
        TRACE_SEND("canEthInit", "CANETH_SOCKET_FAILED");
        closesocket(recvSocket);
        return CANETH_SOCKET_FAILED;
    }
    int broadcast = 1; // Enable option
    if (setsockopt(sendSocket, SOL_SOCKET, SO_BROADCAST, (char *)&broadcast, sizeof(broadcast)) == -1) // Allow broadcasting on this socket
    {
        TRACE_SEND("canEthInit", "CANETH_SOCKET_FAILED");
        closesocket(recvSocket);
        closesocket(sendSocket);
        return CANETH_SOCKET_FAILED;
    }

    //
    // Init critical sections
    //
    for (int i = 0; i < MAX_CAN_GATEWAYS; i++)
        canBufArr[i].init();

    //
    // Create worker thread
    //
#ifdef _WIN32   
    recvThreadHdl = CreateThread(NULL, 0, canRecvThread, 0, 0, NULL);
    if (recvThreadHdl == 0)
#else  
    if (pthread_create(&recvThreadHdl, NULL, canRecvThread, NULL) == -1)
#endif   
    {
        TRACE_SEND("canEthInit", "CANETH_CREATE_THREAD_FAILED");
        closesocket(recvSocket);
        closesocket(sendSocket);
        for (int i = 0; i < MAX_CAN_GATEWAYS; i++)
            canBufArr[i].destroy();
        return CANETH_CREATE_THREAD_FAILED;
    }

    return CANETH_SUCCESS;
}


/**
 * Shutdown the CAN-ETH Ethernet interface.
 *
 * The background receive thread is terminated and all CAN channels closed.
 */
void canEthShutdown()
{
    if (recvThreadHdl != 0)
    {
#ifdef _WIN32      
        TerminateThread(recvThreadHdl, 0);
#else      
        pthread_cancel(recvThreadHdl);
#endif      
        recvThreadHdl = 0;
        closesocket(recvSocket);
        closesocket(sendSocket);

        for (int i = 0; i < MAX_CAN_GATEWAYS; i++)
            canBufArr[i].destroy();
   }
#ifdef _WIN32   
   WSACleanup();
#endif   
}


/**
 * Open a CAN channel to a remote CAN-ETH gateway
 *
 * @param hostName String with IP address or host name of the
 *                 remote CAN-ETH gateway.
 *
 * @return Handle to CAN channel or negative error code
 */
CanHandle canEthOpen(const char *const hostName)
{
    if (recvThreadHdl == 0)
    {
        TRACE_SEND("canEthOpen", "CANETH_ILLEGAL_STATE_ERROR");
        return CANETH_ILLEGAL_STATE_ERROR;
    }

    //
    // Determine host address
    //
    sockaddr_in hostAddr;
    hostAddr.sin_family = AF_INET;
    hostAddr.sin_port = htons(UDP_PORT);
    hostAddr.sin_addr.s_addr = inet_addr((char *) hostName);
    if (hostAddr.sin_addr.s_addr == INADDR_NONE)
    {
        struct hostent *hostInfo;

        hostInfo = gethostbyname((char *) hostName);
        if (hostInfo == NULL)
        {
            TRACE_SEND("canEthOpen", "CANETH_INVALID_HOST");
            return CANETH_INVALID_HOST;
        }
        hostAddr.sin_addr = *(struct in_addr *) hostInfo->h_addr;
    }

    //
    // Find free CAN communication channel
    //
    for (int i = 0; i < MAX_CAN_GATEWAYS; i++)
    {
        if (canBufArr[i].size == 0)
        {
            //
            // Init receive buffer
            //
            canBufArr[i].lock();
            canBufArr[i].readIdx = 0;
            canBufArr[i].writeIdx = 0;
            canBufArr[i].gatewayIpAddr = hostAddr;
            canBufArr[i].recvNotifyFunc = NULL;
            canBufArr[i].size = canBufArr[i].CAN_BUF_SIZE;
            canBufArr[i].unlock();

            return i;
        }
    }
    TRACE_SEND("canEthOpen", "CANETH_MAX_OPEN");
    return CANETH_MAX_OPEN;
}


/**
 * Install a receive notification handler.
 *
 * Must be called after canEthOpen
 *
 * @param hdl    Handle to identify CAN channel
 * @param func   Pointer to callback function
 *
 * @return CANETH_SUCCESS or a negative error code
 */
int canEthInstallRecvCallBack(CanHandle hdl, void (*func) (CanHandle hdl))
{
    if (hdl >= MAX_CAN_GATEWAYS)
    {
        TRACE_SEND("canEthInstallRecvCallBack", "CANETH_ILLEGAL_ARGUMENT_ERROR");
        return CANETH_ILLEGAL_ARGUMENT_ERROR;
    }
    if (canBufArr[hdl].size == 0)
    {
        TRACE_SEND("canEthInstallRecvCallBack", "CANETH_ILLEGAL_STATE_ERROR");
        return CANETH_ILLEGAL_STATE_ERROR;
    }

    canBufArr[hdl].lock();
    canBufArr[hdl].recvNotifyFunc = func;
    canBufArr[hdl].unlock();

    return CANETH_SUCCESS;
}


/**
 * Close a CAN receive channel
 *
 * @param hdl    Handle to identify CAN channel
 */
void canEthClose(CanHandle hdl)
{
    if (hdl < MAX_CAN_GATEWAYS)
    {
        canBufArr[hdl].lock();
        canBufArr[hdl].size = 0;
        canBufArr[hdl].unlock();
    }
}


/**
 * Send a CAN message
 *
 * @param hdl    Handle to identify CAN channel
 * @param msg    Pointer to CAN message to be sent
 *
 * @return CANETH_SUCCESS if sent
 */
int canEthSendMsg(CanHandle hdl, const CanMsg *const msg)
{
    int rc;
    CanEthFrame frame;

    if (hdl >= MAX_CAN_GATEWAYS)
    {
        TRACE_SEND("canEthSendMsg", "CANETH_ILLEGAL_ARGUMENT_ERROR");
        return CANETH_ILLEGAL_ARGUMENT_ERROR;
    }
    if (canBufArr[hdl].size == 0)
    {
        TRACE_SEND("canEthSendMsg", "CANETH_ILLEGAL_STATE_ERROR");
        return CANETH_ILLEGAL_STATE_ERROR;
    }

    // Pack message into a frame
    canEthFrameInit(&frame);
    int len = canEthFramePack(&frame, msg);

    //
    // Send Ethernet frame
    //
    rc = sendto(sendSocket, (char *)&frame, len, 0, (struct sockaddr *) &canBufArr[hdl].gatewayIpAddr, sizeof(canBufArr[hdl].gatewayIpAddr));
    if (rc == -1)
    {
        TRACE_SEND("canEthSendMsg", "CANETH_SEND_FAILED");
        return CANETH_SEND_FAILED;
    }
#ifdef TRACE_TOOL
    else
    {
        char buf[1000];
        sprintf(buf, "dlc=%d flags=%d/%d %lX:%02X %02X %02X %02X %02X %02X %02X %02X",
            msg->len, msg->ext, msg->rtr, msg->id,
            (unsigned int)msg->data[0], (unsigned int)msg->data[1],
            (unsigned int)msg->data[2], (unsigned int)msg->data[3],
            (unsigned int)msg->data[4], (unsigned int)msg->data[5],
            (unsigned int)msg->data[6], (unsigned int)msg->data[7]);
        TTrace::Debug()->Send("canEthSendMsg", buf);
    }
#endif

    return CANETH_SUCCESS;
}


/**
 * Retrieves a CAN message from the receive buffer. If no
 * messages if available, CANETH_NO_MSG is returned.
 *
 * @param hdl    Handle to identify CAN channel
 * @param msg    Pointer to container for received CAN message
 *
 * @return CANETH_SUCCESS if OK or CANETH_NO_MSG if no message
 *         received.
 */
int canEthRecv(CanHandle hdl, CanMsg *msg)
{
    if (hdl >= MAX_CAN_GATEWAYS)
    {
      TRACE_SEND("canEthRecv", "CANETH_ILLEGAL_ARGUMENT_ERROR");
        return CANETH_ILLEGAL_ARGUMENT_ERROR;
    }
    if (canBufArr[hdl].size == 0)
    {
        TRACE_SEND("canEthRecv", "CANETH_ILLEGAL_STATE_ERROR");
        return CANETH_ILLEGAL_STATE_ERROR;
    }

    canBufArr[hdl].lock();
    if (canBufArr[hdl].readIdx == canBufArr[hdl].writeIdx)
    {
        canBufArr[hdl].unlock();
        return CANETH_NO_MSG;
    }
    *msg = canBufArr[hdl].data[canBufArr[hdl].readIdx];
    canBufArr[hdl].readIdx = (canBufArr[hdl].readIdx + 1) % canBufArr[hdl].size;
    canBufArr[hdl].unlock();

#ifdef TRACE_TOOL
    char buf[1000];
    sprintf(buf, "dlc=%d flags=%d/%d %lX:%02X %02X %02X %02X %02X %02X %02X %02X",
        msg->len, msg->ext, msg->rtr, msg->id,
        (unsigned int) msg->data[0], (unsigned int) msg->data[1],
        (unsigned int) msg->data[2], (unsigned int) msg->data[3],
        (unsigned int) msg->data[4], (unsigned int) msg->data[5],
        (unsigned int) msg->data[6], (unsigned int) msg->data[7]);
    TRACE_SEND("canEthRecv", buf);
#endif

   return CANETH_SUCCESS;
}
